# pyre-unsafe

from typing import Any

def date(value, arg: Any = ...) -> Any: ...
def escapejs(value: str) -> str: ...
def linebreaks(value: str, autoescape: bool = ...) -> str: ...
def slugify(value: str) -> str: ...
