# pyre-unsafe

from typing import Any, Dict, Optional, Union

class BaseCookie(dict):
    def __init__(self, input: Optional[Union[str, Dict[str, str]]] = None) -> None: ...

class SimpleCookie(BaseCookie):
    def __getitem__(self, key: str) -> Any: ...
    def update(self, other: Any) -> None: ...
    def get(self, key: Any) -> Any: ...
    def load(self, vals: Dict[Any, Any]) -> None: ...
