from typing import Any

def new(
    nbits,
    prefix: Any = ...,
    suffix: Any = ...,
    initial_value: int = ...,
    overflow: int = ...,
    little_endian: bool = ...,
    allow_wraparound: bool = ...,
    disable_shortcut: bool = ...,
): ...
