# pyre-unsafe

import datetime
from typing import Dict

from django.db.backends.utils import CursorWrapper

class BaseDatabaseWrapper:
    alias: str
    settings_dict: Dict[str, object]
    def __init__(self, settings_dict, alias=...): ...
    def cursor(self) -> CursorWrapper: ...
    def close(self): ...
    def close_if_unusable_or_obsolete(self): ...
    def connection(self) -> ...: ...
