# pyre-unsafe

from typing import List

from django.http import HttpRequest

DEBUG: int = ...
INFO: int = ...
SUCCESS: int = ...
WARNING: int = ...
ERROR: int = ...

def add_message(request: HttpRequest, int, str) -> None: ...
def get_messages(request: HttpRequest) -> List[str]: ...
