# pyre-unsafe

class CursorWrapper:
    def __init__(self, cursor, db): ...
    def __enter__(self) -> "CursorWrapper": ...
    def execute(self, sql, params=None): ...
    def executemany(self, sql, param_list): ...
    def fetchone(self): ...
